@@ -18,7 +18,6 @@ import edu.uic.ncdm.venn.data.VennData;
 
 import java.util.HashMap;
 import java.util.Iterator;
-import java.util.Random;
 import java.util.Set;
 
 public class VennAnalytic {
@@ -35,12 +34,13 @@ public class VennAnalytic {
     private String[] circleLabels;
     private double[][] centers;
     private double[] diameters;
+    private double stepsize;
     private double maxArea;
     private int totalCount;
 
     public VennAnalytic() {
-        minStress = .0001;
-        stress = 1.0;
+        stepsize = .01;
+        minStress = .000001;
     }
 
     public VennDiagram compute(VennData vd) {
@@ -54,11 +54,20 @@ public class VennAnalytic {
             processElementData(data);
 
         computeInitialConfiguration();
-
         scaleDiameters(isAreas);
+        scaleConfiguration();
 
-        if (nCircles > 1)
-            estimatePositions();
+        minimizeGlobal();
+
+        double[][] previousCenters = new double[nCircles][2];
+        copyCircles(centers, previousCenters);
+        double previousStress = stress;
+
+        minimizeLocal();
+        
+        if (stress > previousStress)
+            copyCircles(previousCenters, centers);
+        stress = computeStress();
 
         return collectResults();
     }
@@ -85,6 +94,7 @@ public class VennAnalytic {
             }
             s = s.substring(0, s.length() - 1);
             residualLabels[i - 1] = s;
+            // System.out.println(residualLabels[i - 1] + " " + polyAreas[i] + " " + polyData[i]);
         }
         // System.out.println("stress = " + stress + ", stress01 = " + stress01 + ", stress05 = " + stress05);
 
@@ -207,7 +217,7 @@ public class VennAnalytic {
         return Integer.parseInt(b, 2);
     }
 
-    private void renderVenn() {
+    private void calculateAreas() {
         totalCount = 0;
         int size = 200;
         byte[][][] bis = new byte[nCircles][size][size];
@@ -261,7 +271,7 @@ public class VennAnalytic {
 
     public void computeInitialConfiguration() {
         if (nCircles < 3) {
-            randomStart();
+            fixedStart();
             return;
         }
 
@@ -289,7 +299,7 @@ public class VennAnalytic {
         }
 
         if (max[0] == min[0] || max[1] == min[1]) {
-            randomStart();
+            fixedStart();
             return;
         }
 
@@ -299,11 +309,13 @@ public class VennAnalytic {
         }
     }
 
-    private void randomStart() {
-        Random random = new Random(13579);
+    private void fixedStart() {
+        double theta = Math.PI / 2;
+        double delta = 2 * Math.PI / nCircles;
         for (int i = 0; i < nCircles; i++) {
-            centers[i][0] = .25 + random.nextDouble() / 2;
-            centers[i][1] = .25 + random.nextDouble() / 2;
+            centers[i][0] = .5 + Math.cos(theta);
+            centers[i][1] = .5 + Math.sin(theta);
+            theta -= delta;
         }
     }
 
@@ -324,7 +336,7 @@ public class VennAnalytic {
         }
         for (int i = 1; i < s.length; i++) {
             for (int j = 0; j < i; j++) {
-                s[i][j] = - s[i][j];
+                s[i][j] = -s[i][j];
                 s[j][i] = s[i][j];
             }
         }
@@ -375,8 +387,42 @@ public class VennAnalytic {
         }
     }
 
-    private double computeLoss() {
+    private void scaleDiameters(boolean isAreas) {
+        diameters = new double[nCircles];
+        double area;
+        for (int j = 0; j < nCircles; j++) {
+            if (isAreas)
+                area = circleData[j] / maxArea;
+            else
+                area = circleData[j] / nTot;
+            diameters[j] = 2 * Math.sqrt(area / Math.PI / nCircles);
+        }
+    }
+
+    private void scaleConfiguration() {
+        double vc = 0;
+        for (int j = 0; j < 2; j++) {
+            double mc = 0;
+            for (int k = 0; k < nCircles; k++)
+                mc += centers[k][j];
+            mc /= nCircles;
+            for (int k = 0; k < nCircles; k++) {
+                centers[k][j] -= mc;
+                vc += centers[k][j] * centers[k][j];
+            }
+        }
+        vc = 10. * Math.sqrt(vc / (2 * nCircles));
+        if (vc > 0) {
+            for (int j = 0; j < 2; j++) {
+                for (int k = 0; k < nCircles; k++)
+                    centers[k][j] /= vc;
+            }
+        }
+    }
+
+    private double computeStress() {
         /* regression through origin */
+        calculateAreas();
         double xx = 0;
         double xy = 0;
         int n = polyData.length;
@@ -401,65 +447,92 @@ public class VennAnalytic {
         return sse / sst;
     }
 
-    private void scaleDiameters(boolean isAreas) {
-        diameters = new double[nCircles];
-        double area;
-        double maxDiameter = Double.NEGATIVE_INFINITY;
-        for (int j = 0; j < nCircles; j++) {
-            if (isAreas)
-                area = circleData[j] / maxArea;
-            else
-                area = (double) circleData[j] / nTot;
-            diameters[j] = 2. * Math.sqrt(area / Math.PI / nCircles);
-            maxDiameter = Math.max(diameters[j], maxDiameter);
-        }
-        if (maxDiameter > .2) {
-            for (int j = 0; j < nCircles; j++)
-                diameters[j] *= .2 / maxDiameter;
-        }
-    }
-
-    private void estimatePositions() {
-        double stepsize = .01;
-        double[][] bestCenters = new double[nCircles][2];
-        int worse = 0;
-        for (int iter = 0; iter < 500; iter++) {
+    private void minimizeGlobal() {
+        double lastStress = 1;
+        for (int iter = 0; iter < 100; iter++) {
             recenter();
-            renderVenn();
-            double s = computeLoss();
-            if (s < stress) {
-                stress = s;
-                copyCenters(bestCenters);
-                worse = 0;
-            } else {
-                worse++;
-            }
-            moveCenters(stepsize);
-            // System.out.println("iteration, loss " + iter + " " + stress);
-            if (stress < minStress || worse > 50)
+            stress = computeStress();
+            moveGlobal();
+            // System.out.println("global iteration, loss " + iter + " " + stress);
+            if (stress < minStress || lastStress - stress < minStress)
                 break;
-        }
-        for (int i = 0; i < nCircles; i++) {
-            centers[i][0] = bestCenters[i][0];
-            centers[i][1] = bestCenters[i][1];
+            lastStress = stress;
         }
         recenter();
-        renderVenn();
-        computeLoss();
+        stress = computeStress();
     }
 
-    private void moveCenters(double stepsize) {
-        double[][] gradients = computeGradients(stepsize);
+    private void minimizeLocal() {
+        double lastStress = 1;
+        for (int iter = 0; iter < 100; iter++) {
+            recenter();
+            stress = computeStress();
+            moveLocal();
+            // System.out.println("local iteration, loss " + iter + " " + stress);
+            if (stress < minStress || lastStress - stress < minStress)
+                break;
+            lastStress = stress;
+        }
+        recenter();
+        stress = computeStress();
+    }
+
+    private void moveGlobal() {
+        double[][] gradients = new double[nCircles][2];
+        for (int i = 0; i < nPolygons; i++) {
+            String s = Integer.toBinaryString(i);
+            char[] c = s.toCharArray();
+            int offset = nCircles - c.length;
+            for (int j = 0; j < c.length; j++) {
+                if (c[j] == '0')
+                    continue;
+                int jo = j + offset;
+                for (int k = j + 1; k < c.length; k++) {
+                    if (c[k] == '0')
+                        continue;
+                    int ko = k + offset;
+                    double resid = polyAreas[i] - polyHats[i];
+                    double dx = resid * stepsize * (centers[jo][0] - centers[ko][0]);
+                    double dy = resid * stepsize * (centers[jo][1] - centers[ko][1]);
+                    gradients[jo][0] += dx;
+                    gradients[jo][1] += dy;
+                    gradients[ko][0] -= dx;
+                    gradients[ko][1] -= dy;
+                }
+            }
+        }
         for (int i = 0; i < nCircles; i++) {
             centers[i][0] += gradients[i][0];
             centers[i][1] += gradients[i][1];
         }
     }
 
-    private void copyCenters(double[][] bestCenters) {
+    private void moveLocal() {
+        double[][] gradients = new double[nCircles][2];
         for (int i = 0; i < nCircles; i++) {
-            bestCenters[i][0] = centers[i][0];
-            bestCenters[i][1] = centers[i][1];
+            centers[i][0] += stepsize;
+            double xPlus = computeStress();
+            centers[i][0] -= 2 * stepsize;
+            double xMinus = computeStress();
+            centers[i][0] += stepsize;
+            if (xPlus < xMinus)
+                gradients[i][0] = stepsize;
+            else
+                gradients[i][0] = -stepsize;
+
+            centers[i][1] += stepsize;
+            double yPlus = computeStress();
+            centers[i][1] -= 2 * stepsize;
+            double yMinus = computeStress();
+            centers[i][1] += stepsize;
+            if (yPlus < yMinus)
+                gradients[i][1] = stepsize;
+            else
+                gradients[i][1] = -stepsize;
+        }
+        for (int i = 0; i < nCircles; i++) {
+            centers[i][0] += gradients[i][0];
+            centers[i][1] += gradients[i][1];
         }
     }
 
@@ -475,31 +548,11 @@ public class VennAnalytic {
         for (int i = 0; i < nCircles; i++) {
             centers[i][0] = .5 + centers[i][0] - cx;
             centers[i][1] = .5 + centers[i][1] - cy;
-            centers[i][0] = Math.max(Math.min(centers[i][0], 1 - diameters[i] / 2), diameters[i] / 2);
-            centers[i][1] = Math.max(Math.min(centers[i][1], 1 - diameters[i] / 2), diameters[i] / 2);
         }
     }
 
-    private double[][] computeGradients(double stepsize) {
-        double[][] gradients = new double[nCircles][2];
-        for (int i = 0; i < nPolygons; i++) {
-            char[] c = encode(i);
-            for (int j = 0; j < c.length; j++) {
-                if (c[j] == '0')
-                    continue;
-                for (int k = j + 1; k < c.length; k++) {
-                    if (c[k] == '0')
-                        continue;
-                    double resid = polyAreas[i] - polyHats[i];
-                    double dx = resid * stepsize * (centers[j][0] - centers[k][0]);
-                    double dy = resid * stepsize * (centers[j][1] - centers[k][1]);
-                    gradients[j][0] += dx;
-                    gradients[j][1] += dy;
-                    gradients[k][0] -= dx;
-                    gradients[k][1] -= dy;
-                }
-            }
-        }
-        return gradients;
+    private void copyCircles(double[][] a, double[][] b) {
+        for (int i = 0; i < nCircles; i++)
+            System.arraycopy(a[i], 0, b[i], 0, 2);
     }
-}
\ No newline at end of file
+}
